"use strict";
var _a;
Object.defineProperty(exports, "__esModule", { value: true });
const utils_1 = require("@hibas123/utils");
const types_1 = require("./types");
const nanoid = require("nanoid/generate");
const isEqual = require("lodash.isequal");
const ALPHABET = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz";
class Snapshot {
    constructor() {
        this.id = nanoid(ALPHABET, 21);
        this.isSubscribed = false;
    }
}
const QueryRequest = Symbol("queryRequest");
const Snapshots = Symbol("snapshots");
class OfflineError extends Error {
    constructor() {
        super("Offline");
    }
}
exports.OfflineError = OfflineError;
class AwaitStore {
    constructor(_value) {
        this._value = _value;
        this.observable = new utils_1.Observable();
        this.subscribe = this.subscribe.bind(this);
        this.unsubscribe = this.unsubscribe.bind(this);
    }
    get value() {
        return this._value;
    }
    send(value) {
        this._value = value;
        this.observable.send(value);
    }
    subscribe(handler) {
        this.observable.subscribe(handler);
    }
    unsubscribe(handler) {
        this.observable.unsubscribe(handler);
    }
    awaitValue(val) {
        let ignore;
        let prms = new Promise(yes => {
            const cb = () => {
                if (this._value === val) {
                    yes();
                    this.observable.unsubscribe(cb);
                    return true;
                }
                return false;
            };
            ignore = () => {
                this.observable.unsubscribe(cb);
            };
            if (!cb()) {
                this.observable.subscribe(cb);
            }
        });
        return {
            then: prms.then.bind(prms),
            catch: prms.catch.bind(prms),
            ignore: () => ignore()
        };
    }
    close() {
        this.observable.close();
    }
}
exports.AwaitStore = AwaitStore;
// TODO: Request Timeout
class Database {
    constructor(url, database, accesskey, authkey, rootkey) {
        this.offlineObersvable = new AwaitStore(true);
        this.requests = new Map();
        this[_a] = new Map();
        const u = new URL(url);
        if (authkey)
            u.searchParams.set("authkey", authkey);
        if (rootkey)
            u.searchParams.set("rootkey", rootkey);
        if (accesskey)
            u.searchParams.set("accesskey", accesskey);
        u.searchParams.set("database", database);
        this.url = u;
        this._connect();
    }
    get offline() {
        return {
            subscribe: this.offlineObersvable.subscribe,
            unsubscribe: this.offlineObersvable.unsubscribe
        };
    }
    _connect() {
        this.socket = new Database.WebSocket(this.url.href);
        this.socket.onopen = () => {
            this.offlineObersvable.send(false);
            this[Snapshots].forEach(ss => !ss.isSubscribed ? ss.resubscribe() : undefined);
        };
        this.socket.onerror = err => { console.error(err); };
        this.socket.onmessage = msg => {
            let message = JSON.parse(msg.data);
            switch (message.ns) {
                case "message":
                    this.handleResponse(message.data);
                    break;
                case "snapshot":
                    this.handleSnapshot(message.data);
                    break;
                case "error_msg":
                    // TODO: Some kind of error handling etc.
                    console.error("Server error:", message.data);
                    break;
            }
        };
        this.socket.onclose = () => {
            this[Snapshots].forEach(ss => {
                ss.isSubscribed = false;
            });
            this.offlineObersvable.send(true);
            setTimeout(() => {
                this._connect();
            }, 3000);
        };
    }
    handleSnapshot({ id, data }) {
        let handler = this[Snapshots].get(id);
        if (handler)
            handler.receivedData(data);
    }
    handleResponse({ id, error, data }) {
        let req = this.requests.get(id);
        if (req) {
            if (error)
                req(new Error(data), undefined);
            else
                req(undefined, data);
        }
        else {
            console.error("Invalid response", id);
        }
    }
    async [(_a = Snapshots, QueryRequest)](query, id, ns = "v2") {
        if (!id)
            id = nanoid(ALPHABET, 21);
        return new Promise(async (yes, no) => {
            const finish = (err, data) => {
                clearTimeout(to);
                err ? no(err) : yes(data);
                if (typeof offlinePrms !== "undefined")
                    offlinePrms.ignore();
                this.requests.delete(id);
            };
            let to = setTimeout(() => {
                finish(new Error("Timeout!"), undefined);
            }, 15000);
            await this.offlineObersvable.awaitValue(false);
            let offlinePrms = this.offlineObersvable.awaitValue(true);
            offlinePrms.then(() => {
                finish(new OfflineError(), undefined);
            });
            this.requests.set(id, finish);
            this.socket.send(JSON.stringify({ ns, data: { id, query } }));
        });
    }
    collection(name) {
        return new CollectionRef(this, [name]);
    }
    async collections() {
        return this[QueryRequest]({
            type: "list",
            path: []
        }).then(res => res.map(elm => new CollectionRef(this, elm)));
    }
    close() {
        this.socket.close();
        this.offlineObersvable.close();
    }
}
exports.default = Database;
class Query {
    constructor(db, prefix) {
        this.db = db;
        this.prefix = prefix;
        this._where = [];
    }
    _getNew() {
        let q = new Query(this.db, this.prefix);
        q._where = [...this._where];
        q._limit = this._limit;
        q._orderBy = this._orderBy;
        return q;
    }
    _getDocRef(id) {
        return new DocumentRef(this.db, [...this.prefix, id || nanoid(ALPHABET, 32)]);
    }
    get database() {
        return this.db;
    }
    get path() {
        return [...this.prefix];
    }
    where(fieldPath, opStr, value) {
        let q = this._getNew();
        q._where.push({ fieldPath, opStr, value });
        return q;
    }
    orderBy(fieldPath, direction) {
        let q = this._getNew();
        q._orderBy = { fieldPath, direction };
        return q;
    }
    /**
     * WIP: the functionality is not stable yet!
     * @param limit The number of items
     */
    limit(limit) {
        let q = this._getNew();
        q._limit = limit;
        return q;
    }
    _getField(data, fieldpath) {
        let fields = fieldpath.split(".");
        let value = data;
        while (fields.length > 0) {
            if (!value)
                break;
            value = value[fields.shift()];
        }
        return value;
    }
    _order(data) {
        if (!this._orderBy)
            return data;
        data = data.sort((a, b) => {
            let val1 = this._getField(a.data(), this._orderBy.fieldPath);
            let val2 = this._getField(b.data(), this._orderBy.fieldPath);
            if (["number", "bigint", "boolean", "string"].indexOf(typeof val1) < 0)
                return 1;
            if (["number", "bigint", "boolean", "string"].indexOf(typeof val2) < 0)
                return 1;
            //TODO: Check how to resolve error without transpiler hacks
            //@ts-ignore
            return ((val1 > val2) - (val2 > val1));
        });
        if (this._orderBy.direction === "desc")
            data = data.reverse();
        return data;
    }
    async get() {
        let data = this._order((await this.db[QueryRequest]({
            type: "get",
            path: this.prefix,
            options: {
                limit: this._limit,
                where: this._where
            }
        })).map(e => {
            return {
                id: e.id,
                data: () => e.data,
                ref: this._getDocRef(e.id)
            };
        }));
        return {
            docs: data,
            docChanges: () => {
                return data.map(doc => {
                    return {
                        doc,
                        type: "added"
                    };
                });
            },
            size: data.length,
            empty: data.length <= 0
        };
    }
    async onSnapshot(next) {
        return new Query.QuerySnapshot(this, next).subscribe();
    }
}
Query.QuerySnapshot = class extends Snapshot {
    constructor(query, next) {
        super();
        this.query = query;
        this.next = next;
        this.stored = new Map();
    }
    async subscribe() {
        this.query.database[Snapshots].set(this.id, this);
        await this.resubscribe();
        return () => this.unsubscribe();
    }
    unsubscribe() {
        this.query.database[Snapshots].delete(this.id);
        return this.query.database[QueryRequest]({
            type: "unsubscribe",
            path: this.query.path
        }, this.id, "unsubscribe");
    }
    _change(changes) {
        let docs = Array.from(this.stored.values());
        docs = this.query._order(docs); //TODO: a little inefficient...
        this.next(undefined, {
            docs,
            docChanges: () => changes,
            empty: docs.length <= 0,
            size: docs.length
        });
    }
    receivedData(data) {
        let changes = data.map(raw => {
            let doc = {
                id: raw.id,
                data: () => raw.data,
                ref: this.query._getDocRef(raw.id)
            };
            if (raw.type === "deleted")
                this.stored.delete(raw.id);
            else
                this.stored.set(raw.id, doc);
            return {
                doc,
                type: raw.type
            };
        });
        this._change(changes);
    }
    resubscribe() {
        if (this.isSubscribed)
            return;
        this.isSubscribed = true;
        return this.query.database[QueryRequest]({
            type: "snapshot",
            options: {
                where: this.query._where,
                limit: this.query._limit
            },
            path: this.query.path
        }, this.id, "snapshot").catch(err => {
            this.query.database[Snapshots].delete(this.id);
            return Promise.reject(err);
        }).then((data) => {
            let newStored = new Map();
            let changes = data.map(e => {
                let change = {
                    doc: {
                        id: e.id,
                        data: () => e.data,
                        ref: this.query._getDocRef(e.id)
                    },
                    type: "added"
                };
                newStored.set(change.doc.id, change.doc);
                let old = this.stored.get(e.id);
                if (old) {
                    this.stored.delete(change.doc.id);
                    if (isEqual(old, e.data)) {
                        return undefined;
                    }
                    else {
                        change.type = "modified";
                    }
                }
                return change;
            }).filter(e => !!e);
            changes = [...changes, ...Array.from(this.stored.values()).map(stored => {
                    stored.data = () => null;
                    return {
                        doc: stored,
                        type: "deleted"
                    };
                })];
            this.stored = newStored;
            this._change(changes);
        }).catch(err => {
            if (!(err instanceof OfflineError)) {
                this.next(err, undefined);
            }
            this.isSubscribed = false;
        });
    }
    destroy() {
        if (this.query.database[Snapshots].has(this.id)) {
            this.unsubscribe();
        }
        this.next = undefined;
        this.query = undefined;
        this.stored = undefined;
    }
};
class CollectionRef extends Query {
    get name() {
        return this.prefix[this.prefix.length - 1];
    }
    doc(id) {
        return this._getDocRef(id);
    }
    keys() {
        return this.db[QueryRequest]({
            type: "keys",
            path: this.prefix
        });
    }
    async add(data) {
        let id = await this.db[QueryRequest]({
            type: "add",
            path: this.prefix,
            data
        });
        return new DocumentRef(this.db, [...this.prefix, id]);
    }
    delete() {
        return this.db[QueryRequest]({
            type: "delete-collection",
            path: this.prefix
        });
    }
}
class DocumentRef {
    constructor(db, prefix) {
        this.db = db;
        this.prefix = prefix;
    }
    get id() {
        return this.prefix[this.prefix.length - 1];
    }
    get database() {
        return this.db;
    }
    get path() {
        return [...this.prefix];
    }
    get() {
        return this.db[QueryRequest]({
            type: "get",
            path: this.prefix
        });
    }
    set(data) {
        return this.db[QueryRequest]({
            type: "set",
            path: this.prefix,
            data
        });
    }
    update(update) {
        let u = {};
        for (const field in update) {
            const data = update[field];
            let res;
            if (typeof data === "object") {
                if (data[types_1.TimestampSymbol]) {
                    res = {
                        type: "timestamp"
                    };
                }
                else if (data[types_1.PushSymbol]) {
                    res = {
                        type: "push",
                        value: data.value
                    };
                }
                else if (data[types_1.IncrementSymbol]) {
                    res = {
                        type: "increment",
                        value: data.value
                    };
                }
                else {
                    res = {
                        type: "value",
                        value: data
                    };
                }
            }
            else {
                res = {
                    type: "value",
                    value: data
                };
            }
            u[field] = res;
        }
        return this.db[QueryRequest]({
            type: "update",
            path: this.prefix,
            data: u
        });
    }
    delete() {
        return this.db[QueryRequest]({
            type: "delete",
            path: this.prefix
        });
    }
    async onSnapshot(next) {
        return new DocumentRef.DocumentSnapshot(this, next).subscribe();
    }
    collection(name) {
        return new CollectionRef(this.db, [...this.prefix, name]);
    }
}
DocumentRef.DocumentSnapshot = class extends Snapshot {
    constructor(document, next) {
        super();
        this.document = document;
        this.next = next;
        this.stored = undefined;
    }
    async subscribe() {
        this.document.database[Snapshots].set(this.id, this);
        await this.resubscribe();
        return () => this.unsubscribe();
    }
    unsubscribe() {
        this.document.database[Snapshots].delete(this.id);
        return this.document.database[QueryRequest]({
            type: "unsubscribe",
            path: this.document.path
        }, this.id, "unsubscribe");
    }
    _change() {
        this.next(undefined, {
            id: this.document.id,
            data: () => this.stored,
            ref: this.document
        });
    }
    receivedData(data) {
        this.stored = data.data;
        this._change();
    }
    resubscribe() {
        if (this.isSubscribed)
            return;
        this.isSubscribed = true;
        return this.document.database[QueryRequest]({
            type: "snapshot",
            path: this.document.path
        }, this.id, "snapshot").catch(err => {
            this.document.database[Snapshots].delete(this.id);
            return Promise.reject(err);
        }).then((data) => {
            this.isSubscribed = true;
            if (!isEqual(this.stored, data)) {
                this.stored = data;
                this._change();
            }
        }).catch(err => {
            if (!(err instanceof OfflineError)) {
                this.next(err, undefined);
            }
            this.isSubscribed = false;
        });
    }
    destroy() {
        if (this.document.database[Snapshots].has(this.id)) {
            this.unsubscribe();
        }
        this.next = undefined;
        this.document = undefined;
        this.stored = undefined;
    }
};
//# sourceMappingURL=database.js.map