import { Observable, ObserverCallback } from "@hibas123/utils";
import { DocumentChangeType, DocumentUpdate, FieldPath, ICollectionRef, IDocumentChange, IDocumentRef, IDocumentSnapshot, IDocumentSnapshotHandler, IncrementSymbol, IQuery, IQuerySnapshot, IQuerySnapshotHandler, OrderByDirection, PushSymbol, TimestampSymbol, WhereFilterOp } from "./types";
import nanoid = require("nanoid/generate");
import isEqual = require("lodash.isequal");


const ALPHABET = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz";

interface Change {
   data: any;
   document: string;
   type: DocumentChangeType;
}

type IWriteQueries = "set" | "update" | "delete" | "add";
type ICollectionQueries = "get" | "add" | "keys" | "delete-collection" | "list";
type IDocumentQueries = "get" | "set" | "update" | "delete";

type ICallQueryTypes = IWriteQueries | ICollectionQueries | IDocumentQueries | "snapshot" | "unsubscribe";

interface ITypedQuery<T> {
   path: string[];
   type: T;
   data?: any;
   options?: any;
}

type DocRes = {
   id: string,
   data: any
}

abstract class Snapshot<T> {
   protected id: string = nanoid(ALPHABET, 21);
   public isSubscribed: boolean = false;

   abstract subscribe(): Promise<() => void>;
   abstract unsubscribe(): void;
   abstract receivedData(data: T): void;
   abstract resubscribe(): void;
   abstract destroy(): void;

}

const QueryRequest = Symbol("queryRequest");
const Snapshots = Symbol("snapshots");

export class OfflineError extends Error {
   constructor() {
      super("Offline");
   }
}

export class AwaitStore<T> {
   private observable = new Observable<T>();
   constructor(private _value: T) {
      this.subscribe = this.subscribe.bind(this);
      this.unsubscribe = this.unsubscribe.bind(this);
   }

   get value() {
      return this._value;
   }

   send(value: T) {
      this._value = value;
      this.observable.send(value);
   }

   subscribe(handler: ObserverCallback<T>) {
      this.observable.subscribe(handler);
   }

   unsubscribe(handler: ObserverCallback<T>) {
      this.observable.unsubscribe(handler);
   }

   awaitValue(val: T): PromiseLike<void> & { catch: (cb: (err: any) => PromiseLike<void>) => PromiseLike<void>, ignore: () => void } {

      let ignore;

      let prms = new Promise(yes => {
         const cb = () => {
            if (this._value === val) {
               yes();
               this.observable.unsubscribe(cb);
               return true;
            }

            return false
         }

         ignore = () => {
            this.observable.unsubscribe(cb);
         }

         if (!cb()) {
            this.observable.subscribe(cb);
         }
      });

      return {
         then: prms.then.bind(prms),
         catch: prms.catch.bind(prms),
         ignore: () => ignore()
      }
   }

   close() {
      this.observable.close();
   }
}

// TODO: Request Timeout
export default class Database {
   static WebSocket: typeof WebSocket;

   private offlineObersvable = new AwaitStore(true);

   public get offline() {
      return {
         subscribe: this.offlineObersvable.subscribe,
         unsubscribe: this.offlineObersvable.unsubscribe
      }
   }

   private socket: WebSocket;
   private requests = new Map<string, (err: Error, data: any) => void>();

   private url: URL;
   constructor(url: string, database: string, accesskey?: string, authkey?: string, rootkey?: string) {
      const u = new URL(url);
      if (authkey)
         u.searchParams.set("authkey", authkey);
      if (rootkey)
         u.searchParams.set("rootkey", rootkey);
      if (accesskey)
         u.searchParams.set("accesskey", accesskey);

      u.searchParams.set("database", database);
      this.url = u;

      this._connect();
   }

   private _connect() {
      this.socket = new Database.WebSocket(this.url.href);

      this.socket.onopen = () => {
         this.offlineObersvable.send(false);
         this[Snapshots].forEach(ss => !ss.isSubscribed ? ss.resubscribe() : undefined);
      };

      this.socket.onerror = err => { console.error(err); };
      this.socket.onmessage = msg => {
         let message: { ns: string, data: any } = JSON.parse(msg.data as string);
         switch (message.ns) {
            case "message":
               this.handleResponse(message.data);
               break;
            case "snapshot":
               this.handleSnapshot(message.data);
               break;
            case "error_msg":
               // TODO: Some kind of error handling etc.
               console.error("Server error:", message.data);
               break;
         }
      }

      this.socket.onclose = () => {
         this[Snapshots].forEach(ss => {
            ss.isSubscribed = false;
         });
         this.offlineObersvable.send(true);
         setTimeout(() => {
            this._connect();
         }, 3000);
      }
   }


   [Snapshots] = new Map<string, Snapshot<any>>();

   private handleSnapshot({ id, data }: { id: string, data: any }) {
      let handler = this[Snapshots].get(id);
      if (handler)
         handler.receivedData(data)
   }

   private handleResponse({ id, error, data }: { id: string, error: boolean, data: string }) {
      let req = this.requests.get(id);
      if (req) {
         if (error)
            req(new Error(data), undefined);
         else
            req(undefined, data);
      } else {
         console.error("Invalid response", id);
      }
   }

   public async [QueryRequest]<T = any>(query?: ITypedQuery<ICallQueryTypes>, id?: string, ns: string = "v2"): Promise<T> {
      if (!id)
         id = nanoid(ALPHABET, 21);
      return new Promise<any>(async (yes, no) => {
         const finish = (err: Error, data: any) => {
            clearTimeout(to);
            err ? no(err) : yes(data);
            if (typeof offlinePrms !== "undefined")
               offlinePrms.ignore();
            this.requests.delete(id);
         }

         let to = setTimeout(() => {
            finish(new Error("Timeout!"), undefined);
         }, 15000);

         await this.offlineObersvable.awaitValue(false);

         let offlinePrms = this.offlineObersvable.awaitValue(true);
         offlinePrms.then(() => {
            finish(new OfflineError(), undefined);
         })

         this.requests.set(id, finish);

         this.socket.send(JSON.stringify({ ns, data: { id, query } }))
      })
   }

   public collection<T = any>(name: string): ICollectionRef<T> {
      return new CollectionRef(this, [name]);
   }

   public async collections(): Promise<ICollectionRef[]> {
      return this[QueryRequest]({
         type: "list",
         path: []
      }).then(res => res.map(elm => new CollectionRef(this, elm)));
   }

   public close() {
      this.socket.close();
      this.offlineObersvable.close();
   }
}

class Query implements IQuery {
   private _limit: number;
   private _where: {
      fieldPath: FieldPath,
      opStr: WhereFilterOp,
      value: any
   }[] = [];
   private _orderBy: {
      fieldPath: FieldPath,
      direction: OrderByDirection
   }

   private _getNew() {
      let q = new Query(this.db, this.prefix);

      q._where = [...this._where];
      q._limit = this._limit;
      q._orderBy = this._orderBy;

      return q;
   }

   protected _getDocRef(id?: string): IDocumentRef<any> {
      return new DocumentRef(this.db, [...this.prefix, id || nanoid(ALPHABET, 32)]);
   }

   public get database() {
      return this.db;
   }

   public get path() {
      return [...this.prefix];
   }

   constructor(protected db: Database, protected prefix: string[]) { }

   public where(fieldPath: FieldPath, opStr: WhereFilterOp, value: any): IQuery {
      let q = this._getNew();
      q._where.push({ fieldPath, opStr, value })
      return q;
   }

   public orderBy(fieldPath: FieldPath, direction: OrderByDirection) {
      let q = this._getNew();
      q._orderBy = { fieldPath, direction }
      return q;
   }


   /**
    * WIP: the functionality is not stable yet!
    * @param limit The number of items
    */
   public limit(limit: number): IQuery {
      let q = this._getNew();
      q._limit = limit;
      return q;
   }

   private _getField(data: any, fieldpath: FieldPath) {
      let fields = fieldpath.split(".");
      let value = data;

      while (fields.length > 0) {
         if (!value)
            break;
         value = value[fields.shift()];
      }

      return value;
   }

   private _order(data: IDocumentSnapshot[]) {
      if (!this._orderBy)
         return data;

      data = data.sort((a, b) => {
         let val1 = this._getField(a.data(), this._orderBy.fieldPath);
         let val2 = this._getField(b.data(), this._orderBy.fieldPath);

         if (["number", "bigint", "boolean", "string"].indexOf(typeof val1) < 0)
            return 1;

         if (["number", "bigint", "boolean", "string"].indexOf(typeof val2) < 0)
            return 1;

         //TODO: Check how to resolve error without transpiler hacks

         //@ts-ignore
         return ((val1 > val2) - (val2 > val1));
      })
      if (this._orderBy.direction === "desc")
         data = data.reverse();
      return data;
   }

   public async get(): Promise<IQuerySnapshot> {
      let data = this._order((await this.db[QueryRequest]<DocRes[]>({
         type: "get",
         path: this.prefix,
         options: {
            limit: this._limit,
            where: this._where
         }
      })).map(e => {
         return <IDocumentSnapshot>{
            id: e.id,
            data: () => e.data,
            ref: this._getDocRef(e.id)
         }
      }))

      return {
         docs: data,
         docChanges: () => {
            return data.map(doc => {
               return <IDocumentChange>{
                  doc,
                  type: "added"
               }
            })
         },
         size: data.length,
         empty: data.length <= 0
      }
   }

   static QuerySnapshot = class extends Snapshot<any> {
      private stored = new Map<string, IDocumentSnapshot>();
      constructor(private query: Query, private next: IQuerySnapshotHandler) {
         super();
      }

      async subscribe() {
         this.query.database[Snapshots].set(this.id, this);
         await this.resubscribe();
         return () => this.unsubscribe();
      }

      public unsubscribe() {
         this.query.database[Snapshots].delete(this.id);
         return this.query.database[QueryRequest]({
            type: "unsubscribe",
            path: this.query.path
         }, this.id, "unsubscribe")
      }

      private _change(changes: IDocumentChange[]) {
         let docs = Array.from(this.stored.values());
         docs = this.query._order(docs); //TODO: a little inefficient...

         this.next(undefined, {
            docs,
            docChanges: () => changes,
            empty: docs.length <= 0,
            size: docs.length
         })
      }

      receivedData(data: (DocRes & { type: DocumentChangeType })[]) {
         let changes = data.map(raw => {
            let doc = <IDocumentSnapshot>{
               id: raw.id,
               data: () => raw.data,
               ref: this.query._getDocRef(raw.id)
            };
            if (raw.type === "deleted")
               this.stored.delete(raw.id);
            else
               this.stored.set(raw.id, doc);
            return <IDocumentChange>{
               doc,
               type: raw.type
            }
         })
         this._change(changes);
      }


      resubscribe() {
         if (this.isSubscribed)
            return;

         this.isSubscribed = true;
         return this.query.database[QueryRequest]<DocRes[]>({
            type: "snapshot",
            options: {
               where: this.query._where,
               limit: this.query._limit
            },
            path: this.query.path
         }, this.id, "snapshot").catch(err => {
            this.query.database[Snapshots].delete(this.id);
            return Promise.reject(err);
         }).then((data) => {
            let newStored = new Map<string, IDocumentSnapshot>();
            let changes = data.map(e => {
               let change = {
                  doc: {
                     id: e.id,
                     data: () => e.data,
                     ref: this.query._getDocRef(e.id)
                  },
                  type: "added"
               }
               newStored.set(change.doc.id, change.doc);
               let old = this.stored.get(e.id);

               if (old) {
                  this.stored.delete(change.doc.id);
                  if (isEqual(old, e.data)) {
                     return undefined;
                  } else {
                     change.type = "modified";
                  }
               }
               return change as IDocumentChange;
            }).filter(e => !!e);

            changes = [...changes, ...Array.from(this.stored.values()).map(stored => {
               stored.data = () => null;
               return <IDocumentChange>{
                  doc: stored,
                  type: "deleted"
               }
            })];

            this.stored = newStored;

            this._change(changes);
         }).catch(err => {
            if (!(err instanceof OfflineError)) {
               this.next(err, undefined);
            }
            this.isSubscribed = false;
         })
      }

      destroy() {
         if (this.query.database[Snapshots].has(this.id)) {
            this.unsubscribe();
         }
         this.next = undefined;
         this.query = undefined;
         this.stored = undefined;
      }
   }

   async onSnapshot(next: IQuerySnapshotHandler): Promise<() => void> {
      return new Query.QuerySnapshot(this, next).subscribe();
   }
}

class CollectionRef extends Query implements ICollectionRef {
   get name() {
      return this.prefix[this.prefix.length - 1];
   }

   doc(id?: string): IDocumentRef {
      return this._getDocRef(id);
   }

   keys(): Promise<string[]> {
      return this.db[QueryRequest]({
         type: "keys",
         path: this.prefix
      });
   }

   async add<T = any>(data: T): Promise<IDocumentRef<T>> {
      let id = await this.db[QueryRequest]({
         type: "add",
         path: this.prefix,
         data
      });
      return new DocumentRef(this.db, [...this.prefix, id]);
   }

   delete() {
      return this.db[QueryRequest]({
         type: "delete-collection",
         path: this.prefix
      });
   }
}


class DocumentRef implements IDocumentRef {
   constructor(private db: Database, private prefix: string[]) { }

   get id() {
      return this.prefix[this.prefix.length - 1];
   }

   public get database() {
      return this.db;
   }

   public get path() {
      return [...this.prefix];
   }

   get(): Promise<any> {
      return this.db[QueryRequest]({
         type: "get",
         path: this.prefix
      });
   }

   set(data: any): Promise<void> {
      return this.db[QueryRequest]({
         type: "set",
         path: this.prefix,
         data
      });
   }

   update(update: DocumentUpdate): Promise<void> {
      let u = {};
      for (const field in update) {
         const data = update[field];
         let res: { type: string, value?: any };
         if (typeof data === "object") {
            if (data[TimestampSymbol]) {
               res = {
                  type: "timestamp"
               }
            } else if (data[PushSymbol]) {
               res = {
                  type: "push",
                  value: data.value
               }
            } else if (data[IncrementSymbol]) {
               res = {
                  type: "increment",
                  value: data.value
               }
            } else {
               res = {
                  type: "value",
                  value: data
               };
            }
         } else {
            res = {
               type: "value",
               value: data
            };
         }

         u[field] = res;
      }

      return this.db[QueryRequest]({
         type: "update",
         path: this.prefix,
         data: u
      });
   }

   delete(): Promise<void> {
      return this.db[QueryRequest]({
         type: "delete",
         path: this.prefix
      });
   }

   static DocumentSnapshot = class extends Snapshot<any> {
      private stored = undefined;
      constructor(private document: DocumentRef, private next: IDocumentSnapshotHandler) {
         super();
      }

      async subscribe() {
         this.document.database[Snapshots].set(this.id, this);
         await this.resubscribe();
         return () => this.unsubscribe();
      }

      public unsubscribe() {
         this.document.database[Snapshots].delete(this.id);
         return this.document.database[QueryRequest]({
            type: "unsubscribe",
            path: this.document.path
         }, this.id, "unsubscribe")
      }

      private _change() {
         this.next(undefined, {
            id: this.document.id,
            data: () => this.stored,
            ref: this.document
         })
      }

      receivedData(data: Change) {
         this.stored = data.data
         this._change();
      }

      resubscribe() {
         if (this.isSubscribed)
            return;
         this.isSubscribed = true;
         return this.document.database[QueryRequest]<any>({
            type: "snapshot",
            path: this.document.path
         }, this.id, "snapshot").catch(err => {
            this.document.database[Snapshots].delete(this.id);
            return Promise.reject(err);
         }).then((data) => {
            this.isSubscribed = true;
            if (!isEqual(this.stored, data)) {
               this.stored = data;
               this._change();
            }
         }).catch(err => {
            if (!(err instanceof OfflineError)) {
               this.next(err, undefined);
            }
            this.isSubscribed = false;
         })
      }

      destroy() {
         if (this.document.database[Snapshots].has(this.id)) {
            this.unsubscribe()
         }

         this.next = undefined;
         this.document = undefined;
         this.stored = undefined;
      }
   }

   async onSnapshot(next: IDocumentSnapshotHandler<any>) {
      return new DocumentRef.DocumentSnapshot(this, next).subscribe();
   }

   collection(name: string): ICollectionRef {
      return new CollectionRef(this.db, [...this.prefix, name]);
   }
}